"""
.. module:: ristretto

ristretto module
================

This module exports the classes :obj:`~oblivious.ristretto.point` and
:obj:`~oblivious.ristretto.scalar` for representing points and scalars. It
also exports the two wrapper classes/namespaces
:obj:`~oblivious.ristretto.python` and :obj:`~oblivious.ristretto.sodium`
that encapsulate pure-Python and shared/dynamic library variants of the
above (respectively) and also include low-level operations that correspond
more directly to the functions found in the underlying libraries.

* Under all conditions, the wrapper class :obj:`~oblivious.ristretto.python`
  is defined and encapsulates a pure-Python variant of every class exported
  by this module as a whole. It also includes pure-Python variants of low-level
  operations that correspond to functions found in the underlying libraries.

* If a shared/dynamic library instance of the
  `libsodium <https://doc.libsodium.org>`__ library is found on the system
  (and successfully loaded at the time this module is imported) or the
  optional `rbcl <https://pypi.org/project/rbcl>`__ package is installed,
  then the wrapper class :obj:`~oblivious.ristretto.sodium` is defined.
  Otherwise, the exported variable ``sodium`` is assigned ``None``.

* If a dynamic/shared library instance is loaded, all classes exported by
  this module correspond to the variants defined within
  :obj:`~oblivious.ristretto.sodium`. Otherwise, they correspond to the
  variants defined within :obj:`~oblivious.ristretto.python`.

For most users, the classes :obj:`~oblivious.ristretto.point` and
:obj:`~oblivious.ristretto.scalar` should be sufficient. When using the
low-level operations that correspond to a specific implementation (*e.g.*,
:obj:`oblivious.ristretto.sodium.add`), users are responsible for ensuring
that inputs have the type and/or representation appropriate for that
operation.
"""
from __future__ import annotations
from typing import Any, NoReturn, Union, Optional
import doctest
import platform
import os
import hashlib
import ctypes
import ctypes.util
import secrets
import base64
import ge25519

#
# Attempt to load rbcl. If no local libsodium shared/dynamic library file
# is found, only pure-Python implementations of the functions and methods
# will be available.
#

try: # pragma: no cover
    import rbcl # pylint: disable=E0401

    # Add synonyms to deal with variations in capitalization of function names.
    setattr(
        rbcl,
        'crypto_core_ristretto255_scalarbytes',
        lambda: rbcl.crypto_core_ristretto255_SCALARBYTES
    )
    setattr(
        rbcl,
        'crypto_core_ristretto255_bytes',
        lambda: rbcl.crypto_core_ristretto255_BYTES
    )
except: # pylint: disable=W0702 # pragma: no cover
    rbcl = None

#
# Use pure-Python implementations of primitives by default.
#

def _zero(n: bytes) -> bool:
    d = 0
    for b in n:
        d |= b
    return ((d - 1) >> 8) % 2 == 1

_sc25519_is_canonical_L = [ # 2^252+27742317777372353535851937790883648493.
    0xed, 0xd3, 0xf5, 0x5c, 0x1a, 0x63, 0x12, 0x58, 0xd6, 0x9c, 0xf7,
    0xa2, 0xde, 0xf9, 0xde, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10
]

def _sc25519_is_canonical(s: bytes) -> bool:
    """
    Confirm that the bytes-like object represents a canonical
    scalar.
    """
    c = 0
    n = 1
    for i in range(31, -1, -1):
        c |= ((s[i] - _sc25519_is_canonical_L[i]) >> 8) & n
        n &= ((s[i] ^ _sc25519_is_canonical_L[i]) - 1) >> 8
    return c != 0

def _sc25519_mul(a: bytes, b: bytes) -> bytes:
    """
    Multiply the two scalars represented by the bytes-like objects
    """
    (a, b) = (int.from_bytes(a, 'little'), int.from_bytes(b, 'little'))
    return (
        (a * b) % (pow(2, 252) + 27742317777372353535851937790883648493)
    ).to_bytes(32, 'little')

def _sc25519_sqmul(s: bytes, n: int, a: bytes) -> bytes:
    """
    Perform repeated squaring of a scalar for the designated number
    of iterations, then multiply the result by another scalar.
    """
    for _ in range(n):
        s = _sc25519_mul(s, s)
    return _sc25519_mul(s, a)

def _sc25519_invert(s: bytes) -> bytes:
    """
    Invert the scalar represented by the bytes-like object.
    """
    b_10 = _sc25519_mul(s, s)
    b_100 = _sc25519_mul(b_10, b_10)
    b_11 = _sc25519_mul(b_10, s)
    b_101 = _sc25519_mul(b_10, b_11)
    b_111 = _sc25519_mul(b_10, b_101)
    b_1001 = _sc25519_mul(b_10, b_111)
    b_1011 = _sc25519_mul(b_10, b_1001)
    b_1111 = _sc25519_mul(b_100, b_1011)
    recip = _sc25519_mul(b_1111, s)

    recip = _sc25519_sqmul(recip, 123 + 3, b_101)
    recip = _sc25519_sqmul(recip, 2 + 2, b_11)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1111)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1111)
    recip = _sc25519_sqmul(recip, 4, b_1001)
    recip = _sc25519_sqmul(recip, 2, b_11)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1111)
    recip = _sc25519_sqmul(recip, 1 + 3, b_101)
    recip = _sc25519_sqmul(recip, 3 + 3, b_101)
    recip = _sc25519_sqmul(recip, 3, b_111)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1111)
    recip = _sc25519_sqmul(recip, 2 + 3, b_111)
    recip = _sc25519_sqmul(recip, 2 + 2, b_11)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1011)
    recip = _sc25519_sqmul(recip, 2 + 4, b_1011)
    recip = _sc25519_sqmul(recip, 6 + 4, b_1001)
    recip = _sc25519_sqmul(recip, 2 + 2, b_11)
    recip = _sc25519_sqmul(recip, 3 + 2, b_11)
    recip = _sc25519_sqmul(recip, 3 + 2, b_11)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1001)
    recip = _sc25519_sqmul(recip, 1 + 3, b_111)
    recip = _sc25519_sqmul(recip, 2 + 4, b_1111)
    recip = _sc25519_sqmul(recip, 1 + 4, b_1011)
    recip = _sc25519_sqmul(recip, 3, b_101)
    recip = _sc25519_sqmul(recip, 2 + 4, b_1111)
    recip = _sc25519_sqmul(recip, 3, b_101)
    recip = _sc25519_sqmul(recip, 1 + 2, b_11)

    return recip

def _ristretto255_is_canonical(s: bytes) -> bool:
    """
    Confirm that the bytes-like object represents a canonical
    Ristretto point.
    """
    c = ((s[31] & 0x7f) ^ 0x7f) % 256
    for i in range(30, 0, -1):
        c |= (s[i] ^ 0xff) % 256
    c = (c - 1) >> 8
    d = ((0xed - 1 - s[0]) >> 8) % 256
    return (1 - (((c & d) | s[0]) & 1)) == 1

class python:
    """
    Wrapper class for pure-Python implementations of primitive operations.

    This class encapsulates pure-Python variants of all low-level operations
    and of both classes exported by this module:
    :obj:`python.scl <scl>`, :obj:`python.rnd <rnd>`,
    :obj:`python.inv <inv>`, :obj:`python.smu <smu>`,
    :obj:`python.pnt <pnt>`, :obj:`python.bas <bas>`,
    :obj:`python.can <can>`, :obj:`python.mul <mul>`,
    :obj:`python.add <add>`, :obj:`python.sub <sub>`,
    :obj:`python.neg <neg>`,
    :obj:`python.point <oblivious.ristretto.python.point>`, and
    :obj:`python.scalar <oblivious.ristretto.python.scalar>`.
    For example, you can perform addition of points using
    the pure-Python point addition implementation.

    >>> p = python.pnt()
    >>> q = python.pnt()
    >>> python.add(p, q) == python.add(q, p)
    True

    Pure-Python variants of the :obj:`python.point <point>` and
    :obj:`python.scalar <scalar>` classes always employ pure-Python
    implementations of operations when their methods are invoked.

    >>> p = python.point()
    >>> q = python.point()
    >>> p + q == q + p
    True

    Nevertheless, all bytes-like objects, :obj:`point` objects, and
    :obj:`scalar` objects accepted and emitted by the various operations and
    class methods in :obj:`python` are compatible with those accepted and
    emitted by the operations and class methods in :obj:`sodium`.
    """
    @staticmethod
    def pnt(h: bytes = None) -> bytes:
        """
        Return point from 64-byte vector (normally obtained via hashing).

        >>> p = python.pnt(hashlib.sha512('123'.encode()).digest())
        >>> p.hex()
        '047f39a6c6dd156531a25fa605f017d4bec13b0b6c42f0e9b641c8ee73359c5f'
        """
        return ge25519.ge25519_p3.from_hash_ristretto255(
            hashlib.sha512(python.rnd()).digest() if h is None else h
        )

    @staticmethod
    def bas(s: bytes) -> bytes:
        """
        Return base point multiplied by supplied scalar.

        >>> python.bas(scalar.hash('123'.encode())).hex()
        '4c207a5377f3badf358914f20b505cd1e2a6396720a9c240e5aff522e2446005'
        """
        t = bytearray(s)
        t[31] &= 127

        return ge25519.ge25519_p3.scalar_mult_base(t).to_bytes_ristretto255()

    @staticmethod
    def can(p: bytes) -> bytes:
        """
        Normalize the representation of a point into its canonical form.

        >>> p = point.hash('123'.encode())
        >>> python.can(p) == p
        True
        """
        return p # In this module, the canonical representation is used at all times.

    @staticmethod
    def mul(s: bytes, p: bytes) -> bytes:
        """
        Multiply the point by the supplied scalar and return the result.

        >>> p = python.pnt(hashlib.sha512('123'.encode()).digest())
        >>> s = python.scl(bytes.fromhex(
        ...     '35c141f1c2c43543de9d188805a210abca3cd39a1e986304991ceded42b11709'
        ... ))
        >>> python.mul(s, p).hex()
        '183a06e0fe6af5d7913afb40baefc4dd52ae718fee77a3a0af8777c89fe16210'
        """
        p3 = ge25519.ge25519_p3.from_bytes_ristretto255(p)
        if not _ristretto255_is_canonical(p) or p3 is None:
            return bytes(32) # pragma: no cover

        t = bytearray(s)
        t[31] &= 127

        return p3.scalar_mult(t).to_bytes_ristretto255()

    @staticmethod
    def add(p: bytes, q: bytes) -> bytes:
        """
        Return sum of the supplied points.

        >>> p = point.hash('123'.encode())
        >>> q = point.hash('456'.encode())
        >>> python.add(p, q).hex()
        '7076739c9df665d416e68b9512f5513bf1d0181a2aacefdeb1b7244528a4dd77'
        """
        p_p3 = ge25519.ge25519_p3.from_bytes_ristretto255(p)
        q_p3 = ge25519.ge25519_p3.from_bytes_ristretto255(q)
        if (
            not _ristretto255_is_canonical(p) or p_p3 is None or
            not _ristretto255_is_canonical(q) or q_p3 is None
        ):
            return bytes(32) # pragma: no cover

        q_cached = ge25519.ge25519_cached.from_p3(q_p3)
        r_p1p1 = ge25519.ge25519_p1p1.add(p_p3, q_cached)
        r_p3 = ge25519.ge25519_p3.from_p1p1(r_p1p1)
        return r_p3.to_bytes_ristretto255()

    @staticmethod
    def sub(p: bytes, q: bytes) -> bytes:
        """
        Return result of subtracting second point from first point.

        >>> p = point.hash('123'.encode())
        >>> q = point.hash('456'.encode())
        >>> python.sub(p, q).hex()
        '1a3199ca7debfe31a90171696d8bab91b99eb23a541b822a7061b09776e1046c'
        """
        p_p3 = ge25519.ge25519_p3.from_bytes_ristretto255(p)
        q_p3 = ge25519.ge25519_p3.from_bytes_ristretto255(q)
        if (
            not _ristretto255_is_canonical(p) or p_p3 is None or
            not _ristretto255_is_canonical(q) or q_p3 is None
        ):
            return bytes(32) # pragma: no cover

        q_cached = ge25519.ge25519_cached.from_p3(q_p3)
        r_p1p1 = ge25519.ge25519_p1p1.sub(p_p3, q_cached)
        r_p3 = ge25519.ge25519_p3.from_p1p1(r_p1p1)
        return r_p3.to_bytes_ristretto255()

    @staticmethod
    def neg(p: bytes) -> bytes:
        """
        Return the additive inverse of a point.

        >>> p = point.hash('123'.encode())
        >>> q = point.hash('456'.encode())
        >>> python.add(python.neg(p), python.add(p, q)) == q
        True
        """
        return python.sub(bytes(32), p)

    @staticmethod
    def rnd() -> bytes:
        """
        Return random non-zero scalar.

        >>> len(python.rnd())
        32
        """
        while True:
            r = bytearray(secrets.token_bytes(32))
            r[-1] &= 0x1f
            if _sc25519_is_canonical(r) and not _zero(r):
                return r

    @classmethod
    def scl(cls, s: bytes = None) -> Optional[bytes]:
        """
        Return supplied byte vector if it is a valid scalar; otherwise, return
        ``None``. If no byte vector is supplied, return a random scalar.

        >>> s = python.scl()
        >>> t = python.scl(s)
        >>> s == t
        True
        >>> python.scl(bytes([255] * 32)) is None
        True
        """
        if s is None:
            return cls.rnd()

        s = bytearray(s)
        s[-1] &= 0x1f

        return bytes(s) if _sc25519_is_canonical(s) else None

    @staticmethod
    def inv(s: bytes) -> bytes:
        """
        Return the inverse of a scalar (modulo
        ``2**252 + 27742317777372353535851937790883648493``).

        >>> s = python.scl()
        >>> p = python.pnt()
        >>> python.mul(python.inv(s), python.mul(s, p)) == p
        True
        """
        return _sc25519_invert(s)

    @staticmethod
    def smu(s: bytes, t: bytes) -> bytes:
        """
        Return scalar multiplied by another scalar.

        >>> s = python.scl()
        >>> t = python.scl()
        >>> python.smu(s, t) == python.smu(t, s)
        True
        """
        return _sc25519_mul(s, t)

#
# Attempt to load primitives from libsodium, if it is present;
# otherwise, use the rbcl library, if it is present. Otherwise,
# silently assign ``None`` to ``sodium``.
#

try:
    def _call_variant_unwrapped(length, function, x=None, y=None):
        """
        Wrapper to invoke external function.
        """
        buf = ctypes.create_string_buffer(length)
        if y is not None:
            function(buf, x, y)
        elif x is not None:
            function(buf, x)
        else:
            function(buf)
        return buf.raw

    def _call_variant_wrapped(_, function, x=None, y=None): # pragma: no cover
        """
        Wrapper to invoke external (wrapped) function.
        """
        if y is not None:
            return function(x, y)
        if x is not None:
            return function(x)
        return function()

    _sodium = None
    _call_variant = _call_variant_unwrapped

    # Attempt to load libsodium shared/dynamic library file.
    xdll = ctypes.cdll if platform.system() != 'Windows' else ctypes.windll
    libf = ctypes.util.find_library('sodium') or ctypes.util.find_library('libsodium')
    if libf is not None:
        _sodium = xdll.LoadLibrary(libf)
    else: # pragma: no cover
        # Perform explicit search in case `ld` is not present in environment.
        libf = 'libsodium.so' if platform.system() != 'Windows' else 'libsodium.dll'
        for var in ['PATH', 'LD_LIBRARY_PATH']:
            if var in os.environ:
                for path in os.environ[var].split(os.pathsep):
                    try:
                        _sodium = ctypes.cdll.LoadLibrary(path + os.path.sep + libf)
                        break
                    except: # pylint: disable=W0702
                        continue

        # Default to bindings exported by the rbcl library if the above attempts
        # failed and rbcl is available.
        if _sodium is None and rbcl is not None: # pragma: no cover
            _sodium = rbcl
            _call_variant = _call_variant_wrapped

    # Add method variants that are not present in libsodium.
    if _sodium is not rbcl and _sodium is not None: # pragma: no cover

        def _crypto_scalarmult_ristretto255_allow_scalar_zero(buf, s, p):
            """
            Variant of scalar-point multiplication function that permits
            a scalar corresponding to the zero residue.
            """
            r = _sodium.crypto_scalarmult_ristretto255(buf, s, p)

            if (1 - _zero(s)) * int(r == -1):
                raise RuntimeError('libsodium error (possibly due to invalid input)')

            return buf

        def _crypto_scalarmult_ristretto255_base_allow_scalar_zero(buf, s):
            """
            Variant of scalar-point multiplication function that permits
            a scalar corresponding to the zero residue.
            """
            r = _sodium.crypto_scalarmult_ristretto255_base(buf, s)

            if (1 - _zero(s)) * int(r == -1):
                raise RuntimeError('libsodium error (possibly due to invalid input)')

            return buf

        setattr(
            _sodium,
            'crypto_scalarmult_ristretto255_allow_scalar_zero',
            _crypto_scalarmult_ristretto255_allow_scalar_zero
        )
        setattr(
            _sodium,
            'crypto_scalarmult_ristretto255_base_allow_scalar_zero',
            _crypto_scalarmult_ristretto255_base_allow_scalar_zero
        )

    # Ensure the chosen version of libsodium (or its substitute) has the
    # necessary primitives.
    assert hasattr(_sodium, 'crypto_core_ristretto255_bytes')
    assert hasattr(_sodium, 'crypto_core_ristretto255_scalarbytes')
    assert hasattr(_sodium, 'crypto_core_ristretto255_scalar_random')
    assert hasattr(_sodium, 'crypto_core_ristretto255_scalar_invert')
    assert hasattr(_sodium, 'crypto_core_ristretto255_scalar_mul')
    assert hasattr(_sodium, 'crypto_core_ristretto255_from_hash')
    assert hasattr(_sodium, 'crypto_scalarmult_ristretto255_base')
    assert hasattr(_sodium, 'crypto_scalarmult_ristretto255')
    assert hasattr(_sodium, 'crypto_core_ristretto255_add')
    assert hasattr(_sodium, 'crypto_core_ristretto255_sub')

    # Exported symbol.
    class sodium:
        """
        Wrapper class for binary implementations of primitive operations.

        When this module is imported, it makes a number of attempts to
        locate an instance of the shared/dynamic library file of the
        `libsodium <https://doc.libsodium.org>`__ library on the host
        system. The sequence of attempts is listed below, in order.

        1. It uses ``ctypes.util.find_library`` to look for ``'sodium'`` or
        ``'libsodium'``.

        2. It attempts to find a file ``libsodium.so`` or ``libsodium.dll`` in
           the paths specified by the ``PATH`` and ``LD_LIBRARY_PATH``
           environment variables.

        3. If the `rbcl <https://pypi.org/project/rbcl>`__ package is
           installed, it reverts to the compiled subset of libsodium included
           in that package.

        If all of the above fail, then :obj:`sodium` is assigned the value
        ``None`` and all classes exported by this module default to their
        pure-Python variants (*i.e.*, those encapsulated within :obj:`python`).
        To confirm that a dynamic/shared library *has been found* when this
        module is imported, evaluate the expression ``sodium is not None``.

        If a shared/dynamic library file has been loaded successfully, this
        class encapsulates shared/dynamic library variants of both classes
        exported by this module and of all the underlying low-level operations:
        :obj:`sodium.scl <scl>`, :obj:`sodium.rnd <rnd>`,
        :obj:`sodium.inv <inv>`, :obj:`sodium.smu <smu>`,
        :obj:`sodium.pnt <pnt>`, :obj:`sodium.bas <bas>`,
        :obj:`sodium.can <can>`, :obj:`sodium.mul <mul>`,
        :obj:`sodium.add <add>`, :obj:`sodium.sub <sub>`,
        :obj:`sodium.neg <neg>`,
        :obj:`sodium.point <oblivious.ristretto.sodium.point>`, and
        :obj:`sodium.scalar <oblivious.ristretto.sodium.scalar>`.
        For example, you can perform addition of points using
        the point addition implementation found in the libsodium
        shared/dynamic library found on the host system.

        >>> p = sodium.pnt()
        >>> q = sodium.pnt()
        >>> sodium.add(p, q) == sodium.add(q, p)
        True

        Methods found in the shared/dynamic library variants of the
        :obj:`point` and :obj:`scalar` classes are wrappers for the
        shared/dynamic library implementations of the underlying
        operations.

        >>> p = sodium.point()
        >>> q = sodium.point()
        >>> p + q == q + p
        True

        Nevertheless, all bytes-like objects, :obj:`point` objects, and
        :obj:`scalar` objects accepted and emitted by the various operations
        and class methods in :obj:`sodium` are compatible with those accepted
        and emitted by the operations and class methods in :obj:`python`.
        """
        _lib = _sodium
        _call_unwrapped = _call_variant_unwrapped
        _call_wrapped = _call_variant_wrapped
        _call = _call_variant

        @staticmethod
        def pnt(h: bytes = None) -> bytes:
            """
            Construct a point from its 64-byte vector representation (normally
            obtained via hashing).

            >>> p = sodium.pnt(hashlib.sha512('123'.encode()).digest())
            >>> p.hex()
            '047f39a6c6dd156531a25fa605f017d4bec13b0b6c42f0e9b641c8ee73359c5f'
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_bytes(),
                sodium._lib.crypto_core_ristretto255_from_hash,
                bytes(
                    hashlib.sha512(sodium.rnd()).digest() if h is None else h
                )
            )

        @staticmethod
        def bas(s: bytes) -> bytes:
            """
            Return the base point multiplied by the supplied scalar.

            >>> sodium.bas(scalar.hash('123'.encode())).hex()
            '4c207a5377f3badf358914f20b505cd1e2a6396720a9c240e5aff522e2446005'
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_scalarmult_ristretto255_base_allow_scalar_zero,
                bytes(s)
            )

        @staticmethod
        def can(p: bytes) -> bytes:
            """
            Normalize the representation of a point into its canonical form.

            >>> p = point.hash('123'.encode())
            >>> sodium.can(p) == p
            True
            """
            # In this module, the canonical representation is used at all times.
            return p

        @staticmethod
        def mul(s: bytes, p: bytes) -> bytes:
            """
            Multiply a point by a scalar and return the result.

            >>> p = sodium.pnt(hashlib.sha512('123'.encode()).digest())
            >>> s = sodium.scl(bytes.fromhex(
            ...     '35c141f1c2c43543de9d188805a210abca3cd39a1e986304991ceded42b11709'
            ... ))
            >>> sodium.mul(s, p).hex()
            '183a06e0fe6af5d7913afb40baefc4dd52ae718fee77a3a0af8777c89fe16210'
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_scalarmult_ristretto255_allow_scalar_zero,
                bytes(s), bytes(p)
            )

        @staticmethod
        def add(p: bytes, q: bytes) -> bytes:
            """
            Return the sum of the supplied points.

            >>> p = point.hash('123'.encode())
            >>> q = point.hash('456'.encode())
            >>> sodium.add(p, q).hex()
            '7076739c9df665d416e68b9512f5513bf1d0181a2aacefdeb1b7244528a4dd77'
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_core_ristretto255_add,
                bytes(p), bytes(q)
            )

        @staticmethod
        def sub(p: bytes, q: bytes) -> bytes:
            """
            Return the result of subtracting the right-hand point from the
            left-hand point.

            >>> p = point.hash('123'.encode())
            >>> q = point.hash('456'.encode())
            >>> sodium.sub(p, q).hex()
            '1a3199ca7debfe31a90171696d8bab91b99eb23a541b822a7061b09776e1046c'
            """
            return sodium._call(
                _sodium.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_core_ristretto255_sub,
                bytes(p), bytes(q)
            )

        @staticmethod
        def neg(p: bytes) -> bytes:
            """
            Return the additive inverse of a point.

            >>> p = point.hash('123'.encode())
            >>> q = point.hash('456'.encode())
            >>> sodium.add(sodium.neg(p), sodium.add(p, q)) == q
            True
            """
            return sodium.sub(bytes(32), p)

        @staticmethod
        def rnd() -> bytes:
            """
            Return random non-zero scalar.

            >>> len(sodium.rnd())
            32
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_core_ristretto255_scalar_random
            )

        @classmethod
        def scl(cls, s: bytes = None) -> Optional[bytes]:
            """
            Return supplied byte vector if it is a valid scalar; otherwise,
            return ``None``. If no byte vector is supplied, return a random
            scalar.

            >>> s = sodium.scl()
            >>> t = sodium.scl(s)
            >>> s == t
            True
            >>> sodium.scl(bytes([255] * 32)) is None
            True
            """
            if s is None:
                return cls.rnd()

            s = bytearray(s)
            s[-1] &= 0x1f

            return bytes(s) if _sc25519_is_canonical(s) else None

        @staticmethod
        def inv(s: bytes) -> bytes:
            """
            Return the inverse of a scalar (modulo
            ``2**252 + 27742317777372353535851937790883648493``).

            >>> s = sodium.scl()
            >>> p = sodium.pnt()
            >>> sodium.mul(sodium.inv(s), sodium.mul(s, p)) == p
            True
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_core_ristretto255_scalar_invert,
                bytes(s)
            )

        @staticmethod
        def smu(s: bytes, t: bytes) -> bytes:
            """
            Return the product of two scalars.

            >>> s = sodium.scl()
            >>> t = sodium.scl()
            >>> sodium.smu(s, t) == sodium.smu(t, s)
            True
            """
            return sodium._call(
                sodium._lib.crypto_core_ristretto255_scalarbytes(),
                sodium._lib.crypto_core_ristretto255_scalar_mul,
                bytes(s), bytes(t)
            )

except: # pylint: disable=W0702 # pragma: no cover
    # Exported symbol.
    sodium = None # pragma: no cover

#
# Dedicated point and scalar data structures derived from `bytes`.
#

for _implementation in [python] + ([sodium] if sodium is not None else []):
    # pylint: disable=cell-var-from-loop
    class point(bytes): # pylint: disable=E0102
        """
        Class for representing a point. Because this class is derived from
        :obj:`bytes`, it inherits methods such as :obj:`bytes.hex` and
        :obj:`bytes.fromhex`.

        >>> len(point.random())
        32
        >>> p = point.hash('123'.encode())
        >>> p.hex()
        '047f39a6c6dd156531a25fa605f017d4bec13b0b6c42f0e9b641c8ee73359c5f'
        >>> point.fromhex(p.hex()) == p
        True
        """
        _implementation = _implementation

        @classmethod
        def random(cls) -> point:
            """
            Return random point object.

            >>> len(point.random())
            32
            """
            return bytes.__new__(cls, cls._implementation.pnt())

        @classmethod
        def bytes(cls, bs: bytes) -> point:
            """
            Return the point object obtained by transforming the supplied
            bytes-like object.

            >>> p = point.bytes(hashlib.sha512('123'.encode()).digest())
            >>> p.hex()
            '047f39a6c6dd156531a25fa605f017d4bec13b0b6c42f0e9b641c8ee73359c5f'
            """
            return bytes.__new__(cls, cls._implementation.pnt(bs))

        @classmethod
        def hash(cls, bs: bytes) -> point:
            """
            Return point object by hashing supplied bytes-like object.

            >>> point.hash('123'.encode()).hex()
            '047f39a6c6dd156531a25fa605f017d4bec13b0b6c42f0e9b641c8ee73359c5f'
            """
            return bytes.__new__(cls, cls._implementation.pnt(hashlib.sha512(bs).digest()))

        @classmethod
        def base(cls, s: scalar) -> Optional[point]:
            """
            Return base point multiplied by supplied scalar if the scalar is valid;
            otherwise, return ``None``.

            >>> point.base(scalar.hash('123'.encode())).hex()
            '4c207a5377f3badf358914f20b505cd1e2a6396720a9c240e5aff522e2446005'

            Use of the scalar corresponding to the zero residue is permitted.

            >>> p = point()
            >>> point.base(scalar.from_int(0)) + p == p
            True
            """
            return bytes.__new__(cls, cls._implementation.bas(s))

        @classmethod
        def from_bytes(cls, bs: bytes) -> point:
            """
            Return the instance corresponding to the supplied bytes-like object.

            >>> p = point.bytes(hashlib.sha512('123'.encode()).digest())
            >>> p == point.from_bytes(p.to_bytes())
            True
            """
            return bytes(bs)

        @classmethod
        def from_base64(cls, s: str) -> point:
            """
            Construct an instance from its Base64 UTF-8 string representation.

            >>> point.from_base64('hoVaKq3oIlxEndP2Nqv3Rdbmiu4iinZE6Iwo+kcKAik=').hex()
            '86855a2aade8225c449dd3f636abf745d6e68aee228a7644e88c28fa470a0229'
            """
            return bytes.__new__(cls, base64.standard_b64decode(s))

        def __new__(cls, bs: bytes = None) -> point:
            """
            If a bytes-like object is supplied, return a point object
            corresponding to the supplied bytes-like object (no checking
            is performed to confirm that the bytes-like object is a valid
            point). If no argument is supplied, return a random point
            object.

            >>> bs = bytes.fromhex(
            ...     '86855a2aade8225c449dd3f636abf745d6e68aee228a7644e88c28fa470a0229'
            ... )
            >>> point(bs).hex()
            '86855a2aade8225c449dd3f636abf745d6e68aee228a7644e88c28fa470a0229'
            >>> len(point())
            32
            """
            return bytes.__new__(cls, bs) if bs is not None else cls.random()

        def canonical(self: point) -> point:
            """
            Normalize the representation of this instance into its canonical form.

            >>> p = point.hash('123'.encode())
            >>> p.canonical() == p
            True
            """
            # In this module, the canonical representation is used at all times.
            return self

        def __mul__(self: point, other: Any) -> NoReturn:
            """
            A point cannot be a left-hand argument for a multiplication operation.

            >>> point() * scalar()
            Traceback (most recent call last):
              ...
            TypeError: point must be on right-hand side of multiplication operator
            """
            raise TypeError('point must be on right-hand side of multiplication operator')

        def __rmul__(self: point, other: Any) -> NoReturn:
            """
            This functionality is implemented exclusively in the method
            :obj:`scalar.__mul__`, as that method pre-empts this method
            when the second argument has the correct type (*i.e.*, it is
            a :obj:`scalar` instance). This method is included so that an
            exception can be raised if an incorrect argument is supplied.

            >>> p = point.hash('123'.encode())
            >>> 2 * p
            Traceback (most recent call last):
              ...
            TypeError: point can only be multiplied by a scalar
            """
            raise TypeError('point can only be multiplied by a scalar')

        def __add__(self: point, other: point) -> Optional[point]:
            """
            Return the sum of this instance and another point.

            >>> p = point.hash('123'.encode())
            >>> q = point.hash('456'.encode())
            >>> (p + q).hex()
            '7076739c9df665d416e68b9512f5513bf1d0181a2aacefdeb1b7244528a4dd77'
            >>> p + (q - q) == p
            True
            """
            return self._implementation.point(self._implementation.add(self, other))

        def __sub__(self: point, other: point) -> Optional[point]:
            """
            Return the result of subtracting another point from this instance.

            >>> p = point.hash('123'.encode())
            >>> q = point.hash('456'.encode())
            >>> (p - q).hex()
            '1a3199ca7debfe31a90171696d8bab91b99eb23a541b822a7061b09776e1046c'
            >>> p - p == point.base(scalar.from_int(0))
            True
            """
            return self._implementation.point(self._implementation.sub(self, other))

        def __neg__(self: point) -> point:
            """
            Return the negation of this instance.

            >>> p = point.hash('123'.encode())
            >>> q = point.hash('456'.encode())
            >>> ((p + q) + (-q)) == p
            True
            """
            return (self - self) - self

        def to_bytes(self: point) -> bytes:
            """
            Return the bytes-like object that represents this instance.

            >>> p = point()
            >>> p.to_bytes() == p
            True
            """
            return bytes(self)

        def to_base64(self: point) -> str:
            """
            Return the Base64 UTF-8 string representation of this instance.

            >>> p = point.from_base64('hoVaKq3oIlxEndP2Nqv3Rdbmiu4iinZE6Iwo+kcKAik=')
            >>> p.to_base64()
            'hoVaKq3oIlxEndP2Nqv3Rdbmiu4iinZE6Iwo+kcKAik='
            """
            return base64.standard_b64encode(self).decode('utf-8')

    class scalar(bytes):
        """
        Class for representing a scalar. Because this class is derived from
        :obj:`bytes`, it inherits methods such as :obj:`bytes.hex` and
        :obj:`bytes.fromhex`.

        >>> len(scalar.random())
        32
        >>> s = scalar.hash('123'.encode())
        >>> s.hex()
        'a665a45920422f9d417e4867efdc4fb8a04a1f3fff1fa07e998e86f7f7a27a03'
        >>> scalar.fromhex(s.hex()) == s
        True
        """
        _implementation = _implementation

        @classmethod
        def random(cls) -> scalar:
            """
            Return random non-zero scalar object.

            >>> len(scalar.random())
            32
            """
            return bytes.__new__(cls, cls._implementation.rnd())

        @classmethod
        def bytes(cls, bs: bytes) -> Optional[scalar]:
            """
            Return scalar object obtained by transforming supplied bytes-like
            object if it is possible to do; otherwise, return ``None``.

            >>> s = python.scl()
            >>> t = scalar.bytes(s)
            >>> s.hex() == t.hex()
            True
            """
            s = cls._implementation.scl(bs)
            return bytes.__new__(cls, s) if s is not None else None

        @classmethod
        def hash(cls, bs: bytes) -> scalar:
            """
            Return scalar object by hashing supplied bytes-like object.

            >>> scalar.hash('123'.encode()).hex()
            'a665a45920422f9d417e4867efdc4fb8a04a1f3fff1fa07e998e86f7f7a27a03'
            """
            h = hashlib.sha256(bs).digest()
            s = cls._implementation.scl(h)
            while s is None:
                h = hashlib.sha256(h).digest()
                s = cls._implementation.scl(h)
            return bytes.__new__(cls, s)

        @classmethod
        def from_int(cls, i: int) -> scalar:
            """
            Construct an instance from its integer (*i.e.*, residue) representation.

            >>> p = point()
            >>> zero = scalar.from_int(0)
            >>> zero * p == p - p
            True
            >>> one = scalar.from_int(1)
            >>> one * p == p
            True
            >>> two = scalar.from_int(2)
            >>> two * p == p + p
            True

            Negative integers are supported (and automatically converted into their
            corresponding least nonnegative residues).

            >>> q = point()
            >>> p - p == scalar.from_int(0) * p
            True
            >>> q - p - p == q + (scalar.from_int(-2) * p)
            True
            """
            return bytes.__new__(
                cls,
                (
                    i % (pow(2, 252) + 27742317777372353535851937790883648493)
                ).to_bytes(32, 'little')
            )

        @classmethod
        def from_bytes(cls, bs: bytes) -> Optional[scalar]:
            """
            Return the instance corresponding to the supplied bytes-like object.

            >>> s = python.scl()
            >>> t = scalar.from_bytes(s)
            >>> s.hex() == t.hex()
            True
            """
            return cls(bs)

        @classmethod
        def from_base64(cls, s: str) -> scalar:
            """
            Construct an instance from its Base64 UTF-8 string representation.

            >>> scalar.from_base64('MS0MkTD2kVO+yfXQOGqVE160XuvxMK9fH+0cbtFfJQA=').hex()
            '312d0c9130f69153bec9f5d0386a95135eb45eebf130af5f1fed1c6ed15f2500'
            """
            return bytes.__new__(cls, base64.standard_b64decode(s))

        def __new__(cls, bs: bytes = None) -> scalar:
            """
            If a bytes-like object is supplied, return a scalar object
            corresponding to the supplied bytes-like object (no checking
            is performed to confirm that the bytes-like object is a valid
            scalar). If no argument is supplied, return a random scalar
            object.

            >>> s = python.scl()
            >>> t = scalar(s)
            >>> s.hex() == t.hex()
            True
            >>> len(scalar())
            32
            """
            return bytes.__new__(cls, bs) if bs is not None else cls.random()

        def __invert__(self: scalar) -> scalar:
            """
            Return the inverse of this instance (modulo
            ``2**252 + 27742317777372353535851937790883648493``).

            >>> s = scalar()
            >>> p = point()
            >>> ((~s) * (s * p)) == p
            True

            The scalar corresponding to the zero residue cannot be inverted.

            >>> ~scalar.from_int(0)
            Traceback (most recent call last):
              ...
            ValueError: cannot invert scalar corresponding to zero
            """
            if _zero(self):
                raise ValueError('cannot invert scalar corresponding to zero')

            return self._implementation.scalar(self._implementation.inv(self))

        def __mul__(self: scalar, other: Union[scalar, point]) -> Union[scalar, point]:
            """
            Multiply the supplied scalar or point by this instance.

            >>> p = point.hash('123'.encode())
            >>> s = scalar.hash('456'.encode())
            >>> (s * p).hex()
            'f61b377aa86050aaa88c90f4a4a0f1e36b0000cf46f6a34232c2f1da7a799f16'
            >>> p = point.from_base64('hoVaKq3oIlxEndP2Nqv3Rdbmiu4iinZE6Iwo+kcKAik=')
            >>> s = scalar.from_base64('MS0MkTD2kVO+yfXQOGqVE160XuvxMK9fH+0cbtFfJQA=')
            >>> (s * s).hex()
            'd4aecf034f60edc5cb32cdd5a4be6d069959aa9fd133c51c9dcfd960ee865e0f'
            >>> isinstance(s * s, scalar)
            True
            >>> (s * p).hex()
            '2208082412921a67f42ea399748190d2b889228372509f2f2d9929813d074e1b'
            >>> isinstance(s * p, point)
            True

            Multiplying any point or scalar by the scalar corresponding to the
            zero residue yields the point or scalar corresponding to zero.

            >>> scalar.from_int(0) * point() == p - p
            True
            >>> scalar.from_int(0) * scalar() == scalar.from_int(0)
            True

            Any attempt to multiply a value or object of an incompatible type by this
            instance raises an exception.

            >>> s * 2
            Traceback (most recent call last):
              ...
            TypeError: multiplication by a scalar is defined only for scalars and points
            """
            if (
                isinstance(other, python.scalar) or
                (sodium is not None and isinstance(other, sodium.scalar))
            ):
                return self._implementation.scalar(self._implementation.smu(self, other))

            if (
                isinstance(other, python.point) or
                (sodium is not None and isinstance(other, sodium.point))
            ):
                return self._implementation.point(self._implementation.mul(self, other))

            raise TypeError(
                'multiplication by a scalar is defined only for scalars and points'
            )

        def __rmul__(self: scalar, other: Union[scalar, point]):
            """
            A scalar cannot be on the right-hand side of a non-scalar.

            >>> point() * scalar()
            Traceback (most recent call last):
              ...
            TypeError: point must be on right-hand side of multiplication operator
            """
            raise TypeError(
                'scalar must be on left-hand side of multiplication operator'
            )

        def to_bytes(self: scalar) -> bytes:
            """
            Return the bytes-like object that represents this instance.

            >>> s = scalar()
            >>> s.to_bytes() == s
            True
            """
            return bytes(self)

        def __int__(self: scalar) ->  int:
            """
            Return the integer (*i.e.*, least nonnegative residue) representation
            of this instance.

            >>> s = scalar()
            >>> int(s * (~s))
            1
            """
            return int.from_bytes(self, 'little')

        def to_int(self: scalar) ->  int:
            """
            Return the integer (*i.e.*, least nonnegative residue) representation
            of this instance.

            >>> s = scalar()
            >>> (s * (~s)).to_int()
            1
            """
            return int(self)

        def to_base64(self: scalar) -> str:
            """
            Return the Base64 UTF-8 string representation of this instance.

            >>> s = scalar.from_base64('MS0MkTD2kVO+yfXQOGqVE160XuvxMK9fH+0cbtFfJQA=')
            >>> s.to_base64()
            'MS0MkTD2kVO+yfXQOGqVE160XuvxMK9fH+0cbtFfJQA='
            """
            return base64.standard_b64encode(self).decode('utf-8')

    # Encapsulate classes for this implementation, regardless of which are
    # exported as the unqualified symbols.
    _implementation.point = point
    _implementation.scalar = scalar

# Redefine top-level wrapper classes to ensure that they appear at the end of
# the auto-generated documentation.
python = python # pylint: disable=self-assigning-variable
sodium = sodium # pylint: disable=self-assigning-variable

if __name__ == '__main__':
    doctest.testmod() # pragma: no cover



"""
Test suite containing functional unit tests for the exported primitives and
classes in the :obj:`oblivious.bn254` module, as well as unit tests
confirming algebraic relationships among primitives.
"""
# pylint: disable=missing-function-docstring
from unittest import TestCase
import importlib
import base64
from bitlist import bitlist
from fountains import fountains

try:
    from oblivious import bn254 # pylint: disable=import-error
except: # pylint: disable=bare-except
    # To support generation of reference specifications for unit tests.
    spec = importlib.util.spec_from_file_location("bn254", "src/oblivious/bn254.py")
    bn254 = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(bn254)

# Constants for the number of input-output pairs to include in each test and for
# representation sizes of data structures (in their binary form).
TRIALS_PER_TEST = 16
POINT_HASH_LEN = 64 # Size of raw hash digest required to construct a point.
POINT_LEN = 32 * 3 # Three 32-byte coordinate values (*x*, *y*, and *z* in projective coordinates).
POINT2_LEN = 192
SCALAR_LEN = 32
SCALAR2_LEN = 384

# To simulate an environment in which the mcl library is absent, some tests set
# ``bn254.mcl`` to ``None``; the references below are used for restoration.
mcl_restore = bn254.mcl

def api_functions():
    """
    Low-level functions that should be available to users within each of the two
    namespaces.
    """
    return {
        'pnt', 'bas', 'can', 'ser', 'des', 'mul', 'add', 'sub', 'neg', 'par',
        'rnd', 'scl', 'sse', 'sde', 'inv', 'smu', 'sad', 'ssu', 'sne',
        'pnt2', 'bas2', 'can2', 'ser2', 'des2', 'mul2', 'add2', 'sub2', 'neg2',
        'rnd2', 'scl2', 'sse2', 'sde2', 'inv2', 'smu2', 'sad2'
    }

def api_classes():
    """
    Classes that should be available to users upon module import.
    """
    return {'point', 'scalar', 'point2', 'scalar2'}

class Test_namespace(TestCase):
    """
    Check that namespaces provide access to the expected
    classes and functions.
    """
    def test_init(self):
        init = importlib.import_module('oblivious.__init__')
        self.assertTrue('bn254' in init.__dict__)

    def test_modules(self):
        module = importlib.import_module('oblivious.bn254')
        self.assertTrue('python' in module.__dict__)
        self.assertTrue('mcl' in module.__dict__)
        self.assertTrue(api_classes().issubset(module.__dict__.keys()))

    def test_python(self):
        self.assertTrue(
            (api_functions() - {'par'}).issubset(set(dir(bn254.python)))
        )
        self.assertTrue(api_classes().issubset(set(dir(bn254.python))))

    def test_mcl(self):
        if bn254.mcl is not None:
            self.assertTrue(api_functions().issubset(set(dir(bn254.mcl))))
            self.assertTrue(api_classes().issubset(set(dir(bn254.mcl))))

def check_or_generate_operation(test, fun, lengths, bits):
    """
    This function does either of two things depending on the parameter ``bits``:
    * checks that test inputs drawn from the :obj:`fountains` input bit stream
      produce the bits provided in the reference output bit vector, or
    * generates a reference output bit vector by applying the function
      to the :obj:`fountains` input bit stream.
    """
    def get_bytes(o):
        if type(o) in (bytes, bytearray, bitlist):
            return o

        cls = bn254.python if isinstance(o, bytes) else bn254.mcl

        # Transform points to canonical form.
        cls_str = str(o.__class__)
        if 'point2' in cls_str or 'G2' in cls_str:
            o = cls.can2(o)
        elif 'point' in cls_str or 'G1' in cls_str:
            o = cls.can(o)

        return (
            cls.ser(o)
            if ('point' in str(o.__class__) or 'G' in str(o.__class__)) else
            cls.sse(o)
        )

    fs = fountains( # Generate the input bit stream.
        sum(lengths),
        seed=bytes(0), # This is also the default; explicit for clarity.
        limit=min(TRIALS_PER_TEST, (len(bits) * 4) if bits is not None else 256),
        bits=bits[:(TRIALS_PER_TEST // 4)] if bits is not None else None,
        function=lambda *args, **kwargs: get_bytes(fun(*args, **kwargs))
    )

    if bits is None: # There is no output reference bit vector, so test is not possible.
        return bitlist(list(fs)).hex() # Return reference output bits for test.

    test.assertTrue(all(fs)) # Check that all outputs match.

def mcl_hidden_and_fallback(hidden=False, fallback=False):
    """
    Return binary wrapper class definition that conforms to the
    scenario being tested.
    """
    if hidden:
        bn254.mcl = None
    elif fallback:
        bn254.mcl = mcl_restore
    else:
        bn254.mcl = mcl_restore

def define_classes(cls, hidden=False, fallback=False): # pylint: disable=too-many-statements
    """
    Define and return four classes of unit tests given a wrapper
    class (``python`` or ``mcl``) for primitive operations.
    """
    class Test_primitives(TestCase):
        """
        Direct tests of primitive operators that operate on bytes-like objects.
        """
        def test_pnt(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.pnt, [POINT_HASH_LEN], bits)

        def test_pnt_none(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                self.assertTrue(len(cls.ser(cls.pnt())) == POINT_LEN)

        def test_bas(
                self,
                bits='82d6'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00111111 # Improve chance of testing with a valid scalar.
                bs = bytes(bs)
                s = cls.scl(bs)
                return cls.bas(s) if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_can(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.can(cls.pnt(bs))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_ser(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.ser(cls.pnt(bs))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_des(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.des(cls.ser(cls.pnt(bs)))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_mul(
                self,
                bits='b9e1'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00111111 # Improve chance of testing with a valid scalar.
                bs[SCALAR_LEN - 1] &= 0b00011111
                bs = bytes(bs)
                (s, p) = (cls.scl(bs[:SCALAR_LEN]), cls.pnt(bs[SCALAR_LEN:]))
                return cls.mul(s, p) if (s is not None and p is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN, POINT_HASH_LEN], bits)

        def test_add(
                self,
                bits='40d9'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.pnt(bs[:POINT_HASH_LEN]), cls.pnt(bs[POINT_HASH_LEN:]))
                return cls.add(p1, p2) if (p1 is not None and p2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_sub(
                self,
                bits='71df'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.pnt(bs[:POINT_HASH_LEN]), cls.pnt(bs[POINT_HASH_LEN:]))
                return cls.sub(p1, p2) if (p1 is not None and p2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_neg(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.neg(cls.pnt(bs))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_par(
                self,
                bits='49f2'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.point.hash(bs), cls.point2.hash(bs))
                return cls.sse2(cls.par(p1, p2))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_rnd(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.rnd()
                self.assertTrue(cls.scl(bytes(s)))

        def test_scl(
                self,
                bits='ffff'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00111111 # Improve chances of testing with a valid scalar.
                bs = bytes(bs)
                r = 0x2523648240000001ba344d8000000007ff9f800000000010a10000000000000d
                return bitlist([
                    1
                    if (bool(cls.scl(bs)) == (int.from_bytes(bs, 'little') < r)) is not None else
                    0
                ])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_scl_none(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                bs = bytes([255] * 32) # Representation value above the maximum scalar value.
                s = cls.scl(bs)
                self.assertTrue(s is None)

        def test_sse(
                self,
                bits='ab29'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                s = cls.scl(bytes(bs))
                return cls.sse(s) if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_sde(
                self,
                bits='ab29'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                s = cls.scl(bytes(bs))
                return cls.sde(cls.sse(s)) if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_inv(
                self,
                bits='0aea'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                bs = bytes(bs)
                s = cls.scl(bs)
                return cls.inv(s) if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_smu(
                self,
                bits='dfad'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                bs[SCALAR_LEN - 1] &= 0b00011111
                bs = bytes(bs)
                (s1, s2) = (cls.scl(bs[:SCALAR_LEN]), cls.scl(bs[SCALAR_LEN:]))
                return cls.smu(s1, s2) if (s1 is not None and s2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN, SCALAR_LEN], bits)

        def test_sad(
                self,
                bits='6c5b'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                bs[SCALAR_LEN - 1] &= 0b00011111
                bs = bytes(bs)
                (s1, s2) = (cls.scl(bs[:SCALAR_LEN]), cls.scl(bs[SCALAR_LEN:]))
                return cls.sad(s1, s2) if (s1 is not None and s2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN, SCALAR_LEN], bits)

        def test_ssu(
                self,
                bits='dc34'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                bs[SCALAR_LEN - 1] &= 0b00011111
                bs = bytes(bs)
                (s1, s2) = (cls.scl(bs[:SCALAR_LEN]), cls.scl(bs[SCALAR_LEN:]))
                return cls.ssu(s1, s2) if (s1 is not None and s2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN, SCALAR_LEN], bits)

        def test_sne(
                self,
                bits='66d6'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                bs = bytes(bs)
                s = cls.scl(bs)
                return cls.sne(s) if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_pnt2(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.pnt2, [POINT_HASH_LEN], bits)

        def test_bas2(
                self,
                bits='1ed0'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.bas2(cls.scalar.hash(bs))
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_can2(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.ser2(cls.can2(cls.pnt2(bs)))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_ser2(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.ser2(cls.can2(cls.pnt2(bs)))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_des2(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.des2(cls.ser2(cls.can2(cls.pnt2(bs))))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_mul2(
                self,
                bits='1e5c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s, p) = (cls.scalar.hash(bs[:SCALAR_LEN]), cls.pnt2(bs[SCALAR_LEN:]))
                return cls.mul2(s, p)
            return check_or_generate_operation(self, fun, [SCALAR_LEN, POINT_HASH_LEN], bits)

        def test_add2(
                self,
                bits='424e'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.pnt2(bs[:POINT_HASH_LEN]), cls.pnt2(bs[POINT_HASH_LEN:]))
                return cls.add2(p1, p2)
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_sub2(
                self,
                bits='15c9'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.pnt2(bs[:POINT_HASH_LEN]), cls.pnt2(bs[POINT_HASH_LEN:]))
                return cls.sub2(p1, p2)
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_neg2(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.neg2(cls.pnt2(bs))
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_rnd2(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.rnd2()
                self.assertTrue(cls.scl2(bytes(s)) is not None)

        def test_scl2(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.sse2(cls.scl2(cls.scalar2.hash(bs).to_bytes()))
            return check_or_generate_operation(self, fun, [SCALAR2_LEN], bits)

        def test_sse2(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.sse2(cls.scalar2.hash(bs))
            return check_or_generate_operation(self, fun, [SCALAR2_LEN], bits)

        def test_sde2(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.sse2(cls.sde2(cls.sse2(cls.scalar2.hash(bs))))
            return check_or_generate_operation(self, fun, [SCALAR2_LEN], bits)

        def test_inv2(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.sse2(cls.inv2(cls.scalar2.hash(bs)))
            return check_or_generate_operation(self, fun, [SCALAR2_LEN], bits)

        def test_smu2(
                self,
                bits='ed98'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s1, s2) = (cls.scalar2.hash(bs[:SCALAR2_LEN]), cls.scalar2.hash(bs[SCALAR2_LEN:]))
                return cls.sse2(cls.smu2(s1, s2))
            return check_or_generate_operation(self, fun, [SCALAR2_LEN, SCALAR2_LEN], bits)

        def test_sad2(
                self,
                bits='1d51'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s1, s2) = (cls.scalar2.hash(bs[:SCALAR2_LEN]), cls.scalar2.hash(bs[SCALAR2_LEN:]))
                return cls.sse2(cls.sad2(s1, s2))
            return check_or_generate_operation(self, fun, [SCALAR2_LEN, SCALAR2_LEN], bits)

    class Test_classes(TestCase): # pylint: disable=too-many-public-methods
        """
        Tests of point and scalar wrapper classes and their methods.
        """
        def test_point_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                self.assertTrue(len(cls.point.random()) == POINT_LEN)

        def test_point_bytes(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.point.bytes, [POINT_HASH_LEN], bits)

        def test_point_hash(
                self,
                bits='c8ea'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.point.hash, [POINT_HASH_LEN], bits)

        def test_point_base(
                self,
                bits='ebed'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chance of testing with a valid scalar.
                s = cls.scalar.bytes(bytes(bs))
                return cls.point.base(s) if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_point_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                p = cls.point()
                self.assertEqual(cls.point.from_bytes(p.to_bytes()), p)

        def test_point_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                p = cls.point()
                self.assertEqual(cls.point.fromhex(p.hex()), p)

        def test_point_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                p = cls.point()
                p_b64 = base64.standard_b64encode(p.to_bytes()).decode('utf-8')
                self.assertEqual(p.to_base64(), p_b64)
                self.assertEqual(cls.point.from_base64(p.to_base64()), p)
                self.assertEqual(cls.point.from_base64(p_b64), p)

        def test_point(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                self.assertTrue(len(cls.point()) == POINT_LEN)

        def test_point_canonical(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.point.bytes(bs).canonical()
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_point_scalar_mul(
                self,
                bits='b9e1'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00111111 # Improve chance of testing with a valid scalar.
                bs[SCALAR_LEN - 1] &= 0b00011111
                bs = bytes(bs)
                (s, p) = (cls.scalar.bytes(bs[:SCALAR_LEN]), cls.point.bytes(bs[SCALAR_LEN:]))
                # Below, ``*`` invokes :obj:`scalar.__mul__`, which delegates to :obj:`mul`
                # due to the type of the second argument.
                return s * p if (s is not None and p is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN, POINT_HASH_LEN], bits)

        def test_point_add(
                self,
                bits='40d9'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (
                    cls.point.bytes(bs[:POINT_HASH_LEN]),
                    cls.point.bytes(bs[POINT_HASH_LEN:])
                )
                return p1 + p2 if (p1 is not None and p2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_point_sub(
                self,
                bits='71df'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (
                    cls.point.bytes(bs[:POINT_HASH_LEN]),
                    cls.point.bytes(bs[POINT_HASH_LEN:])
                )
                return p1 - p2 if (p1 is not None and p2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_point_neg(
                self,
                bits='bc3d'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                p = cls.point.bytes(bs)
                return -p if p is not None else bytes([0])
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_point_pair(
                self,
                bits='9cd3'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (
                    cls.point.hash(bs[:POINT_HASH_LEN]),
                    cls.point2.hash(bs[POINT_HASH_LEN:])
                )
                return p1 @ p2
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_scalar_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar.random()
                self.assertTrue(cls.scalar.bytes(bytes(s)) is not None)
                self.assertTrue(len(s) == SCALAR_LEN and len(s.to_bytes()) == SCALAR_LEN)

        def test_scalar_bytes(
                self,
                bits='ab29'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = bytearray(bs)
                bs[-1] &= 0b00011111 # Improve chances of testing with a valid scalar.
                s = cls.scalar.bytes(bytes(bs))
                return s.to_bytes() if s is not None else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_scalar_hash(
                self,
                bits='1c21'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.scalar.hash, [SCALAR_LEN], bits)

        def test_scalar_to_int(
                self,
                bits='6969'
            ):
            def fun(bs):
                s = cls.scalar.hash(bs)
                return abs(s.to_int()).to_bytes(32, 'little')
            return check_or_generate_operation(self, fun, [32], bits)

        def test_scalar_from_int(
                self,
                bits='d27b'
            ):
            def fun(bs):
                s = cls.scalar.from_int(int.from_bytes(bs, 'little'))
                return s if (s is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [32], bits)

        def test_scalar_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar()
                self.assertEqual(cls.scalar.from_bytes(s.to_bytes()), s)

        def test_scalar_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar()
                self.assertEqual(cls.scalar.fromhex(s.hex()), s)

        def test_scalar_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar()
                s_b64 = base64.standard_b64encode(s.to_bytes()).decode('utf-8')
                self.assertEqual(s.to_base64(), s_b64)
                self.assertEqual(cls.scalar.from_base64(s_b64), s)
                self.assertEqual(cls.scalar.from_base64(s.to_base64()), s)

        def test_scalar(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar()
                self.assertTrue(len(s) == SCALAR_LEN and cls.scalar.bytes(bytes(s)) is not None)
                self.assertTrue(len(bytes(s)) == SCALAR_LEN and len(s.to_bytes()) == SCALAR_LEN)

        def test_scalar_invert(
                self,
                bits='c4e3'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                s = cls.scalar.hash(bs)
                return ~s
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_scalar_mul(
                self,
                bits='93d3'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s1, s2) = (cls.scalar.hash(bs[:SCALAR_LEN]), cls.scalar.hash(bs[SCALAR_LEN:]))
                return s1 * s2 if (s1 is not None and s2 is not None) else bytes([0])
            return check_or_generate_operation(self, fun, [SCALAR_LEN, SCALAR_LEN], bits)

        def test_scalar_add(
                self,
                bits='debe'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                s = cls.scalar.hash(bs[SCALAR_LEN:])
                t = cls.scalar.hash(bs[:SCALAR_LEN])
                return s + t
            return check_or_generate_operation(self, fun, [SCALAR_LEN, SCALAR_LEN], bits)

        def test_scalar_sub(
                self,
                bits='7fc9'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                s = cls.scalar.hash(bs[SCALAR_LEN:])
                t = cls.scalar.hash(bs[:SCALAR_LEN])
                return s - t
            return check_or_generate_operation(self, fun, [SCALAR_LEN, SCALAR_LEN], bits)

        def test_scalar_neg(
                self,
                bits='a1de'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                s = cls.scalar.hash(bs)
                return -s
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_point2_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                self.assertTrue(len(cls.point2.random()) == POINT2_LEN)

        def test_point2_bytes(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.point2.bytes, [POINT_HASH_LEN], bits)

        def test_point2_hash(
                self,
                bits='d51b'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.point2.hash, [POINT_HASH_LEN], bits)

        def test_point2_base(
                self,
                bits='1ed0'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.point2.base(cls.scalar.hash(bs))
            return check_or_generate_operation(self, fun, [SCALAR_LEN], bits)

        def test_point2_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                p = cls.point2()
                self.assertEqual(cls.point2.from_bytes(p.to_bytes()), p)

        def test_point2_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                p = cls.point2()
                self.assertEqual(cls.point2.fromhex(p.hex()), p)

        def test_point2_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                p = cls.point2()
                p_b64 = base64.standard_b64encode(p.to_bytes()).decode('utf-8')
                self.assertEqual(p.to_base64(), p_b64)
                self.assertEqual(cls.point2.from_base64(p.to_base64()), p)
                self.assertEqual(cls.point2.from_base64(p_b64), p)

        def test_point2(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                self.assertTrue(len(cls.point2()) == POINT2_LEN)

        def test_point2_canonical(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return cls.point2.bytes(bs).canonical()
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_point2_scalar_mul(
                self,
                bits='1e5c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s, p) = (cls.scalar.hash(bs[:SCALAR_LEN]), cls.point2.bytes(bs[SCALAR_LEN:]))
                # Below, ``*`` invokes :obj:`scalar.__mul__`, which delegates to :obj:`mul`
                # due to the type of the second argument.
                return s * p
            return check_or_generate_operation(self, fun, [SCALAR_LEN, POINT_HASH_LEN], bits)

        def test_point2_add(
                self,
                bits='424e'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.point2.bytes(bs[:POINT_HASH_LEN]), cls.point2.bytes(bs[POINT_HASH_LEN:]))
                return p1 + p2
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_point2_sub(
                self,
                bits='15c9'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (p1, p2) = (cls.point2.bytes(bs[:POINT_HASH_LEN]), cls.point2.bytes(bs[POINT_HASH_LEN:]))
                return p1 - p2
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN, POINT_HASH_LEN], bits)

        def test_point2_neg(
                self,
                bits='d28c'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                return -cls.point2.bytes(bs)
            return check_or_generate_operation(self, fun, [POINT_HASH_LEN], bits)

        def test_scalar2_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar2.random()
                self.assertTrue(cls.scalar2.bytes(s.to_bytes()) is not None)
                self.assertTrue(len(s) == SCALAR2_LEN and len(s.to_bytes()) == SCALAR2_LEN)

        def test_scalar2_bytes(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                bs = cls.scalar2.hash(bs).to_bytes()
                s = cls.scalar2.bytes(bs)
                return s.to_bytes()
            return check_or_generate_operation(self, fun, [SCALAR2_LEN], bits)

        def test_scalar2_hash(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            return check_or_generate_operation(self, cls.scalar2.hash, [SCALAR2_LEN], bits)

        def test_scalar2_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar2()
                self.assertEqual(cls.scalar2.from_bytes(s.to_bytes()), s)

        def test_scalar2_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar2()
                self.assertEqual(cls.scalar2.fromhex(s.hex()), s)

        def test_scalar2_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar2()
                s_b64 = base64.standard_b64encode(s.to_bytes()).decode('utf-8')
                self.assertEqual(s.to_base64(), s_b64)
                self.assertEqual(cls.scalar2.from_base64(s_b64), s)
                self.assertEqual(cls.scalar2.from_base64(s.to_base64()), s)

        def test_scalar2(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for _ in range(TRIALS_PER_TEST):
                s = cls.scalar2()
                self.assertTrue(cls.scalar2(s.to_bytes()) is not None)
                self.assertTrue(len(s) == SCALAR2_LEN and len(s.to_bytes()) == SCALAR2_LEN)

        def test_scalar2_invert(
                self,
                bits='a567'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                s = cls.scalar2.hash(bs)
                return ~s
            return check_or_generate_operation(self, fun, [SCALAR2_LEN], bits)

        def test_scalar2_mul(
                self,
                bits='b0f7'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s1, s2) = (cls.scalar2.hash(bs[:SCALAR_LEN]), cls.scalar2.hash(bs[SCALAR_LEN:]))
                return s1 * s2
            return check_or_generate_operation(self, fun, [SCALAR2_LEN, SCALAR2_LEN], bits)

        def test_scalar2_add(
                self,
                bits='015e'
            ):
            mcl_hidden_and_fallback(hidden, fallback)
            def fun(bs):
                (s1, s2) = (cls.scalar2.hash(bs[:SCALAR_LEN]), cls.scalar2.hash(bs[SCALAR_LEN:]))
                return s1 + s2
            return check_or_generate_operation(self, fun, [SCALAR2_LEN, SCALAR2_LEN], bits)

    class Test_types(TestCase): # pylint: disable=too-many-public-methods
        """
        Tests verifying that methods that should return point and scalar classes
        do indeed return point and scalar objects of the expected types.
        """
        def test_types_point_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            p = cls.point.random()
            self.assertTrue(isinstance(p, cls.point))

        def test_types_point_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN, limit=1)
            p = cls.point.bytes(bs)
            self.assertTrue(isinstance(p, cls.point))

        def test_types_point_hash(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN, limit=1)
            p = cls.point.hash(bs)
            self.assertTrue(isinstance(p, cls.point))

        def test_types_point_base(self):
            mcl_hidden_and_fallback(hidden, fallback)
            p = cls.point.base(cls.scalar.random())
            self.assertTrue(isinstance(p, cls.point))

        def test_types_point_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            bs = cls.point.random().to_bytes()
            self.assertTrue(isinstance(cls.point.from_bytes(bs), cls.point))

        def test_types_point_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.point.random().hex()
            self.assertTrue(isinstance(cls.point.fromhex(s), cls.point))

        def test_types_point_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.point.random().to_base64()
            self.assertTrue(isinstance(cls.point.from_base64(s), cls.point))

        def test_types_point_canonical(self):
            mcl_hidden_and_fallback(hidden, fallback)
            p = cls.point.base(cls.scalar.random())
            self.assertTrue(isinstance(p.canonical(), cls.point))

        def test_types_point_mul(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(SCALAR_LEN + POINT_HASH_LEN, limit=1)
            (s, p) = (cls.scalar.hash(bs[:SCALAR_LEN]), cls.point.hash(bs[SCALAR_LEN:]))
            self.assertTrue(isinstance(s * p, cls.point))

        def test_types_point_add(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=1)
            (p0, p1) = (cls.point.hash(bs[:POINT_HASH_LEN]), cls.point.hash(bs[POINT_HASH_LEN:]))
            self.assertTrue(isinstance(p0 + p1, cls.point))

        def test_types_point_sub(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=1)
            (p0, p1) = (cls.point.hash(bs[:POINT_HASH_LEN]), cls.point.hash(bs[POINT_HASH_LEN:]))
            self.assertTrue(isinstance(p0 - p1, cls.point))

        def test_types_point_neg(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN, limit=1)
            p = cls.point.hash(bs)
            self.assertTrue(isinstance(-p, cls.point))

        def test_types_point_point2_matmul(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=1)
            (p0, p1) = (cls.point.hash(bs[:POINT_HASH_LEN]), cls.point2.hash(bs[POINT_HASH_LEN:]))
            self.assertTrue(isinstance(p0 @ p1, cls.scalar2))

        def test_types_scalar_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            self.assertTrue(isinstance(cls.scalar.random(), cls.scalar))

        def test_types_scalar_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            bs = cls.scalar.random().to_bytes()
            self.assertTrue(isinstance(cls.scalar.bytes(bs), cls.scalar))

        def test_types_scalar_hash(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(SCALAR_LEN, limit=1)
            self.assertTrue(isinstance(cls.scalar.hash(bs), cls.scalar))

        def test_types_scalar_to_int_from_int(self):
            mcl_hidden_and_fallback(hidden, fallback)
            n = cls.scalar.random().to_int()
            self.assertTrue(isinstance(cls.scalar.from_int(n), cls.scalar))

        def test_types_scalar_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            bs = cls.scalar.random().to_bytes()
            self.assertTrue(isinstance(cls.scalar.from_bytes(bs), cls.scalar))

        def test_types_scalar_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar.random().hex()
            self.assertTrue(isinstance(cls.scalar.fromhex(s), cls.scalar))

        def test_types_scalar_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar.random().to_base64()
            self.assertTrue(isinstance(cls.scalar.from_base64(s), cls.scalar))

        def test_types_scalar_invert(self):
            mcl_hidden_and_fallback(hidden, fallback)
            self.assertTrue(isinstance(~cls.scalar.random(), cls.scalar))

        def test_types_scalar_mul_scalar(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (s0, s1) = (cls.scalar.random(), cls.scalar.random())
            self.assertTrue(isinstance(s0 * s1, cls.scalar))

        def test_types_scalar_add_scalar(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (s0, s1) = (cls.scalar.random(), cls.scalar.random())
            self.assertTrue(isinstance(s0 + s1, cls.scalar))

        def test_types_scalar_sub_scalar(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (s0, s1) = (cls.scalar.random(), cls.scalar.random())
            self.assertTrue(isinstance(s0 - s1, cls.scalar))

        def test_types_scalar_neg(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar.random()
            self.assertTrue(isinstance(-s, cls.scalar))

        def test_types_scalar_mul_point(self):
            mcl_hidden_and_fallback(hidden, fallback)
            self.assertTrue(isinstance(cls.scalar() * cls.point(), cls.point))

        def test_types_point2_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            p = cls.point2.random()
            self.assertTrue(isinstance(p, cls.point2))

        def test_types_point2_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN, limit=1)
            p = cls.point2.bytes(bs)
            self.assertTrue(isinstance(p, cls.point2))

        def test_types_point2_hash(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN, limit=1)
            p = cls.point2.hash(bs)
            self.assertTrue(isinstance(p, cls.point2))

        def test_types_point2_base(self):
            mcl_hidden_and_fallback(hidden, fallback)
            p = cls.point2.base(cls.scalar.random())
            self.assertTrue(isinstance(p, cls.point2))

        def test_types_point2_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            bs = cls.point2.random().to_bytes()
            self.assertTrue(isinstance(cls.point2.from_bytes(bs), cls.point2))

        def test_types_point2_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.point2.random().hex()
            self.assertTrue(isinstance(cls.point2.fromhex(s), cls.point2))

        def test_types_point2_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.point2.random().to_base64()
            self.assertTrue(isinstance(cls.point2.from_base64(s), cls.point2))

        def test_types_point2_canonical(self):
            mcl_hidden_and_fallback(hidden, fallback)
            p = cls.point2.base(cls.scalar.random())
            self.assertTrue(isinstance(p.canonical(), cls.point2))

        def test_types_point2_mul(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(SCALAR_LEN + POINT_HASH_LEN, limit=1)
            (s, p) = (cls.scalar.hash(bs[:SCALAR_LEN]), cls.point2.hash(bs[SCALAR_LEN:]))
            self.assertTrue(isinstance(s * p, cls.point2))

        def test_types_point2_add(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=1)
            (p0, p1) = (cls.point2.hash(bs[:POINT_HASH_LEN]), cls.point2.hash(bs[POINT_HASH_LEN:]))
            self.assertTrue(isinstance(p0 + p1, cls.point2))

        def test_types_point2_sub(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=1)
            (p0, p1) = (cls.point2.hash(bs[:POINT_HASH_LEN]), cls.point2.hash(bs[POINT_HASH_LEN:]))
            self.assertTrue(isinstance(p0 - p1, cls.point2))

        def test_types_point2_neg(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(POINT_HASH_LEN, limit=1)
            self.assertTrue(isinstance(-cls.point2.hash(bs), cls.point2))

        def test_types_scalar2_random(self):
            mcl_hidden_and_fallback(hidden, fallback)
            self.assertTrue(isinstance(cls.scalar2.random(), cls.scalar2))

        def test_types_scalar2_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            bs = cls.scalar2.random().to_bytes()
            self.assertTrue(isinstance(cls.scalar2.bytes(bs), cls.scalar2))

        def test_types_scalar2_hash(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (bs,) = fountains(SCALAR2_LEN, limit=1)
            self.assertTrue(isinstance(cls.scalar2.hash(bs), cls.scalar2))

        def test_types_scalar2_to_bytes_from_bytes(self):
            mcl_hidden_and_fallback(hidden, fallback)
            bs = cls.scalar2.random().to_bytes()
            self.assertTrue(isinstance(cls.scalar2.from_bytes(bs), cls.scalar2))

        def test_types_scalar2_hex_fromhex(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar2.random().hex()
            self.assertTrue(isinstance(cls.scalar2.fromhex(s), cls.scalar2))

        def test_types_scalar2_to_base64_from_base64(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar2.random().to_base64()
            self.assertTrue(isinstance(cls.scalar2.from_base64(s), cls.scalar2))

        def test_types_scalar2_invert(self):
            mcl_hidden_and_fallback(hidden, fallback)
            self.assertTrue(isinstance(~cls.scalar2.random(), cls.scalar2))

        def test_types_scalar2_mul_scalar2(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (s0, s1) = (cls.scalar2.random(), cls.scalar2.random())
            self.assertTrue(isinstance(s0 * s1, cls.scalar2))

        def test_types_scalar2_add_scalar2(self):
            mcl_hidden_and_fallback(hidden, fallback)
            (s0, s1) = (cls.scalar2.random(), cls.scalar2.random())
            self.assertTrue(isinstance(s0 + s1, cls.scalar2))

    class Test_algebra(TestCase):
        """
        Tests of common algebraic properties of low-level operations (for all classes).
        """
        def test_algebra_point_add_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(2 * POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point.hash(bs[:POINT_HASH_LEN]),
                    cls.point.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add(p0, p1), cls.add(p1, p0))

        def test_algebra_point_add_sub_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(2 * POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point.hash(bs[:POINT_HASH_LEN]),
                    cls.point.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add(cls.sub(p0, p1), p1), p0)

        def test_algebra_point_add_neg_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(2 * POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point.hash(bs[:POINT_HASH_LEN]),
                    cls.point.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add(cls.neg(p0), cls.add(p0, p1)), p1)

        def test_algebra_point_neg_neg_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                p = cls.point.hash(bs[:POINT_HASH_LEN])
                self.assertEqual(cls.neg(cls.neg(p)), p)

        def test_algebra_point_add_neg_sub_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(2 * POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point.hash(bs[:POINT_HASH_LEN]),
                    cls.point.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add(p0, cls.neg(p1)), cls.sub(p0, p1))

        def test_algebra_point_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                z = cls.point.base(cls.scalar.from_int(0))
                p = cls.point.hash(bs)
                self.assertEqual(cls.add(z, p), p)
                self.assertEqual(cls.add(p, z), p)

        def test_algebra_scalar_inverse_inverse_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN, limit=TRIALS_PER_TEST):
                s = cls.scalar.hash(bs)
                self.assertEqual(cls.inv(cls.inv(s)), s)

        def test_algebra_scalar_inverse_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN, limit=TRIALS_PER_TEST):
                s = cls.scalar.hash(bs)
                self.assertEqual(cls.inv(cls.inv(s)), s)

        def test_algebra_scalar_inverse_mul_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (s, p) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.point.hash(bs[SCALAR_LEN:])
                )
                self.assertEqual(cls.mul(cls.inv(s), cls.mul(s, p)), p)

        def test_algebra_scalar_mul_point_scalar_zero(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                z = cls.point.base(cls.scalar.from_int(0))
                p = cls.point.hash(bs)
                self.assertEqual(cls.mul(cls.scalar.from_int(0), p), z)

        def test_algebra_scalar_mul_point_scalar_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                p = cls.point.hash(bs)
                self.assertEqual(cls.mul(cls.scalar.from_int(1), p), p)

        def test_algebra_scalar_mul_point_mul_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains((2 * SCALAR_LEN) + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (s0, s1, p0) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.scalar.hash(bs[SCALAR_LEN: SCALAR_LEN + SCALAR_LEN]),
                    cls.point.hash(bs[SCALAR_LEN + SCALAR_LEN:])
                )
                self.assertEqual(
                    cls.mul(s0, cls.mul(s1, p0)),
                    cls.mul(s1, cls.mul(s0, p0))
                )

        def test_algebra_scalar_add_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + SCALAR_LEN, limit=TRIALS_PER_TEST):
                (s0, s1) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.scalar.hash(bs[SCALAR_LEN:])
                )
                self.assertEqual(cls.sad(s0, s1), cls.sad(s1, s0))

        def test_algebra_scalar_add_neg_add_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + SCALAR_LEN, limit=TRIALS_PER_TEST):
                (s0, s1) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.scalar.hash(bs[SCALAR_LEN:])
                )
                self.assertEqual(cls.sad(cls.sad(s0, cls.sne(s0)), s1), s1)

        def test_algebra_scalar_mul_point_mul_associate(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + SCALAR_LEN + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (s0, s1, p0) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.scalar.hash(bs[SCALAR_LEN: SCALAR_LEN + SCALAR_LEN]),
                    cls.point.hash(bs[SCALAR_LEN + SCALAR_LEN:])
                )
                self.assertEqual(
                    cls.mul(s0, cls.mul(s1, p0)),
                    cls.mul(cls.smu(s0, s1), p0)
                )

        def test_algebra_scalar_mul_point_add_distribute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + (2 * POINT_HASH_LEN), limit=TRIALS_PER_TEST):
                (s0, p0, p1) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.point.hash(bs[SCALAR_LEN: SCALAR_LEN + POINT_HASH_LEN]),
                    cls.point.hash(bs[SCALAR_LEN + POINT_HASH_LEN:])
                )
                self.assertEqual(
                    cls.add(cls.mul(s0, p0), cls.mul(s0, p1)),
                    cls.mul(s0, cls.add(p0, p1))
                )

        def test_algebra_scalar_mul_scalar_on_right_hand_side_of_non_scalar(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar.random()
            self.assertRaises(TypeError, lambda: bytes([0]) * s)

        def test_algebra_scalar_mul_point_on_left_hand_side(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar.random()
            p = cls.point.hash(bytes(POINT_LEN))
            self.assertRaises(TypeError, lambda: p * s)

        def test_algebra_point2_add_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point2.hash(bs[:POINT_HASH_LEN]),
                    cls.point2.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.can2(cls.add2(p0, p1)), cls.can2(cls.add2(p1, p0)))

        def test_algebra_point2_add_sub_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point2.hash(bs[:POINT_HASH_LEN]),
                    cls.point2.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add2(cls.sub2(p0, p1), p1), p0)

        def test_algebra_point2_add_neg_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(2 * POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point2.hash(bs[:POINT_HASH_LEN]),
                    cls.point2.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add2(cls.neg2(p0), cls.add2(p0, p1)), p1)

        def test_algebra_point2_neg_neg_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                p = cls.point2.hash(bs[:POINT_HASH_LEN])
                self.assertEqual(cls.neg2(cls.neg2(p)), p)

        def test_algebra_point2_add_neg_sub_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(2 * POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (p0, p1) = (
                    cls.point2.hash(bs[:POINT_HASH_LEN]),
                    cls.point2.hash(bs[POINT_HASH_LEN:])
                )
                self.assertEqual(cls.add2(p0, cls.neg2(p1)), cls.sub2(p0, p1))

        def test_algebra_point2_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                z = cls.point2.base(cls.scalar.from_int(0))
                p = cls.point2.hash(bs)
                self.assertEqual(cls.add2(z, p), p)
                self.assertEqual(cls.add2(p, z), p)

        def test_algebra_scalar_mul_point2_mul_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains((2 * SCALAR_LEN) + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (s0, s1, p0) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.scalar.hash(bs[SCALAR_LEN: SCALAR_LEN + SCALAR_LEN]),
                    cls.point2.hash(bs[SCALAR_LEN + SCALAR_LEN:])
                )
                self.assertEqual(
                    cls.mul(s0, cls.mul2(s1, p0)),
                    cls.mul(s1, cls.mul2(s0, p0))
                )

        def test_algebra_scalar_mul_point2_mul_associate(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + SCALAR_LEN + POINT_HASH_LEN, limit=TRIALS_PER_TEST):
                (s0, s1, p0) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.scalar.hash(bs[SCALAR_LEN: SCALAR_LEN + SCALAR_LEN]),
                    cls.point2.hash(bs[SCALAR_LEN + SCALAR_LEN:])
                )
                self.assertEqual(
                    cls.mul2(s0, cls.mul2(s1, p0)),
                    cls.mul2(cls.smu(s0, s1), p0)
                )

        def test_algebra_scalar_mul_point2_add_distribute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN + (2 * POINT_HASH_LEN), limit=TRIALS_PER_TEST):
                (s0, p0, p1) = (
                    cls.scalar.hash(bs[:SCALAR_LEN]),
                    cls.point2.hash(bs[SCALAR_LEN: SCALAR_LEN + POINT_HASH_LEN]),
                    cls.point2.hash(bs[SCALAR_LEN + POINT_HASH_LEN:])
                )
                self.assertEqual(
                    cls.add2(cls.mul2(s0, p0), cls.mul2(s0, p1)),
                    cls.mul2(s0, cls.add2(p0, p1))
                )

        def test_algebra_scalar_mul_point2_on_left_hand_side(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar.random()
            p = cls.point2.hash(bytes(POINT2_LEN))
            self.assertRaises(TypeError, lambda: p * s)

        def test_algebra_scalar2_inverse_inverse_cancel(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR2_LEN, limit=TRIALS_PER_TEST):
                s = cls.scalar2.hash(bs)
                self.assertEqual(cls.inv2(cls.inv2(s)), s)

        def test_algebra_scalar2_inverse_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR_LEN, limit=TRIALS_PER_TEST):
                s = cls.scalar2.hash(bs)
                self.assertEqual(cls.inv2(cls.inv2(s)), s)

        def test_algebra_scalar2_mul_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR2_LEN + SCALAR2_LEN, limit=TRIALS_PER_TEST):
                (s0, s1) = (
                    cls.scalar2.hash(bs[:SCALAR2_LEN]),
                    cls.scalar2.hash(bs[SCALAR2_LEN:])
                )
                self.assertEqual(cls.smu2(s0, s1), cls.smu2(s1, s0))

        def test_algebra_scalar2_mul_inverse_mul_identity(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR2_LEN + SCALAR2_LEN, limit=TRIALS_PER_TEST):
                (s0, s1) = (
                    cls.scalar2.hash(bs[:SCALAR2_LEN]),
                    cls.scalar2.hash(bs[SCALAR2_LEN:])
                )
                self.assertEqual(cls.smu2(cls.smu2(s0, cls.inv2(s0)), s1), s1)

        def test_algebra_scalar2_add_commute(self):
            mcl_hidden_and_fallback(hidden, fallback)
            for bs in fountains(SCALAR2_LEN + SCALAR2_LEN, limit=TRIALS_PER_TEST):
                (s0, s1) = (
                    cls.scalar2.hash(bs[:SCALAR2_LEN]),
                    cls.scalar2.hash(bs[SCALAR2_LEN:])
                )
                self.assertEqual(cls.sad2(s0, s1), cls.sad2(s1, s0))

        def test_algebra_scalar2_mul_scalar_on_right_hand_side_of_non_scalar(self):
            mcl_hidden_and_fallback(hidden, fallback)
            s = cls.scalar2.random()
            self.assertRaises(TypeError, lambda: bytes([0]) * s)

    # Remove tests for functions/methods for which no pure-Python implementation
    # exists.
    if cls == bn254.python:
        delattr(Test_primitives, 'test_par')
        delattr(Test_classes, 'test_point_pair')
        delattr(Test_types, 'test_types_point_point2_matmul')

    return (
        Test_primitives,
        Test_classes,
        Test_types,
        Test_algebra
    )

# The instantiated test classes below are discovered by the testing framework and
# executed in alphabetical order.
(
    Test_primitives_python_no_mcl,
    Test_classes_python_no_mcl,
    Test_types_python_no_mcl,
    Test_algebra_python_no_mcl
) = define_classes(bn254.python, hidden=True)

if bn254.mcl is not None and bn254.mclbn256 is True:
    (
        Test_primitives_mcl_mclbn256_no_mcl,
        Test_classes_mcl_mclbn256_no_mcl,
        Test_types_mcl_mclbn256_no_mcl,
        Test_algebra_mcl_mclbn256_no_mcl
    ) = define_classes(bn254.mcl, fallback=True)

(Test_primitives_python, Test_classes_python, Test_types_python, Test_algebra_python) = \
    define_classes(bn254.python)

if bn254.mcl is not None:
    (Test_primitives_mcl, Test_classes_mcl, Test_types_mcl, Test_algebra_mcl) = \
        define_classes(bn254.mcl)

if __name__ == "__main__":
    # Generate reference bit lists for tests.
    implementation_test_ensembles = (
        [Test_primitives_python(), Test_classes_python()]
        if bn254.mcl is None else
        [Test_primitives_mcl(), Test_classes_mcl()]
    )

    for tests in implementation_test_ensembles:
        print(
            '\nUnit test reference bit vectors for ' +
            tests.__class__.__name__ + ' methods...'
        )
        for m in [m for m in dir(tests) if m.startswith('test_')]:
            method = getattr(tests, m)
            if 'bits' in method.__code__.co_varnames:
                print('* ' + m + ": '" + method(bits=None) + "'")



"""
Python library that serves as an API for common cryptographic primitives
used to implement OPRF, OT, and PSI protocols.

This module gives users direct access to the individual modules, each of
which is dedicated to a specific elliptic curve and its associated classes
and methods.
"""
from oblivious import ristretto
from oblivious import bn254
